/*
 * Copyright (C) 2015-2016  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#ifndef ASYNCFIFO_USE_GLIB

#define TEST_EVENTFD

#ifdef TEST_EVENTFD
# include <poll.h>
# include <sys/eventfd.h>
# include <unistd.h>
#endif

#include "asyncfifo.h"
#include "memutil.h"
#include "queue.h"
#include "thread.h"

#ifndef LOCK
# define LOCK(a) lock_lock(&(a)->lock)
#endif
#ifndef UNLOCK
# define UNLOCK(a) lock_unlock(&(a)->lock)
#endif

struct _AsyncFifo{
    Lock lock;
#ifdef TEST_EVENTFD
    int available_fd;
#else
    Condition available_cond;
#endif
    Queue* queue;
    Condition done_cond;                /**< The queue is emptied. */
};


AsyncFifo* async_fifo_new(){
    AsyncFifo* fifo = u_new(AsyncFifo,1);
    lock_init(&fifo->lock);
#ifdef TEST_EVENTFD
    fifo->available_fd = eventfd(0,EFD_NONBLOCK|EFD_SEMAPHORE);
#else
    condition_init(&fifo->available_cond);
#endif
    fifo->queue = queue_new();
    condition_init(&fifo->done_cond);
    return fifo;
}


void async_fifo_free(AsyncFifo* fifo){
    async_fifo_wait(fifo);
#ifdef TEST_EVENTFD
    close(fifo->available_fd);
#else
    condition_done(&fifo->available_cond);
#endif
    lock_done(&fifo->lock);
    queue_free(fifo->queue);
    condition_done(&fifo->done_cond);
    u_free(fifo);
}


void async_fifo_push(AsyncFifo* fifo,void* data){
    LOCK(fifo);
    queue_push_head(fifo->queue,data);
#ifdef TEST_EVENTFD
    {
        uint64_t value = 1;
        /* force ignore return value */
        (void)(write(fifo->available_fd,&value,sizeof(uint64_t))+1);
    }
#else
    if(queue_length(fifo->queue)<=0)
        condition_signal(&fifo->available_cond);
#endif
    UNLOCK(fifo);
}


void* async_fifo_pop(AsyncFifo* fifo){
    void* data;
#ifdef TEST_EVENTFD
    for(;;){
        struct pollfd fds;
        fds.fd = fifo->available_fd;
        fds.events = POLLIN;
        poll(&fds,1,-1);
        if(fds.revents&POLLIN){
            uint64_t value;
            if(read(fds.fd,&value,sizeof(uint64_t))>0)
                break;
        }
    }
#endif
    LOCK(fifo);
#ifndef TEST_EVENTFD
    while(queue_length(fifo->queue)<=0)
        condition_wait(&fifo->available_cond,&fifo->lock);
#endif
    data = queue_pop_tail(fifo->queue);
    if(queue_length(fifo->queue)<=0)
        condition_broadcast(&fifo->done_cond);
    UNLOCK(fifo);
    return data;
}


void* async_fifo_try_pop(AsyncFifo* fifo){
    void* data;
#ifdef TEST_EVENTFD
    uint64_t value;
    if(read(fifo->available_fd,&value,sizeof(uint64_t))<=0)
        return NULL;
#endif
    LOCK(fifo);
#ifndef TEST_EVENTFD
    if(queue_length(fifo->queue)<=0){
        UNLOCK(fifo);
        return NULL;
    }
#endif
    data = queue_pop_tail(fifo->queue);
    if(queue_length(fifo->queue)<=0)
        condition_broadcast(&fifo->done_cond);
    UNLOCK(fifo);
    return data;
}


#ifdef TEST_EVENTFD
int async_fifo_fd(AsyncFifo* fifo){
    return fifo->available_fd;
}
#endif


void async_fifo_wait(AsyncFifo* fifo){
    LOCK(fifo);
    while(queue_length(fifo->queue)>0){
        condition_wait(&fifo->done_cond,&fifo->lock);
    }
    UNLOCK(fifo);
}

#endif /* USE_GLIB */
