/*
 * Copyright (C) 2015-2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "debug.h"
#include "eventloop.h"
#include "memutil.h"


typedef struct{
    EventloopFdFunc func;
    void* userdata;
    int fd;
    int events;
}FdData;

typedef struct{
    EventloopTimeoutFunc func;
    void* userdata;
#ifndef EVENTLOOP_USE_GLIB
    int timeout;
    long monotime;
#endif
}TimeoutData;

typedef struct{
    DispatchFunc func;
    void* data;
}DispatchData;

typedef struct{
    enum{
        MSG_QUIT,
        MSG_ADD_FD,
        MSG_ADD_TIMEOUT,
        MSG_DISPATCH
    }type;
    union{
        FdData* fd_data;
        TimeoutData* timeout_data;
        DispatchData* dispatch_data;
    };
}Msg;

#ifdef EVENTLOOP_USE_GLIB

#else

#include <errno.h>
#include <sys/select.h>
#include <time.h>

#include "asyncfifo.h"
#include "list.h"
#include "queue.h"
#include "thread.h"


struct _Eventloop{
    AsyncFifo* msg_fifo;
    List* fd_list;
    Queue* timeout_queue;
    Thread run_thread;
    int quit;
};


static long get_monotime_now(){
    struct timespec monotime;
    clock_gettime(CLOCK_MONOTONIC,&monotime);
    return monotime.tv_sec*1000+monotime.tv_nsec/1000000;
}


static void eventloop_insert_timeout_data(Eventloop* loop,TimeoutData* d){
    List* iter;
    for(iter=loop->timeout_queue->head;;iter=iter->next){
        if(!iter){
            queue_append(loop->timeout_queue,d);
            break;
        }else if(((TimeoutData*)iter->data)->monotime > d->monotime){
            queue_insert_before(loop->timeout_queue,iter,d);
            break;
        }
    }
    DEBUG_p("%d timeouts in queue",queue_length(loop->timeout_queue));
}


static void eventloop_handle_msg(Eventloop* loop,Msg* msg){
    switch(msg->type){
    case MSG_QUIT:
        loop->quit = 1;
        break;
    case MSG_ADD_FD:
        loop->fd_list = list_prepend(loop->fd_list,msg->fd_data);
        break;
    case MSG_ADD_TIMEOUT:
        eventloop_insert_timeout_data(loop,msg->timeout_data);
        break;
    case MSG_DISPATCH:
        msg->dispatch_data->func(msg->dispatch_data->data);
        break;
    }
    u_free(msg);
}


static int eventloop_new_cb(Eventloop* loop){
    for(;;){
        Msg* msg = (Msg*)async_fifo_try_pop(loop->msg_fifo);
        if(!msg)
            return 1;
        eventloop_handle_msg(loop,msg);
    }
}
Eventloop* eventloop_new(){
    Eventloop* loop = u_new(Eventloop,1);
    //loop->wakeup_fd = eventfd(0,EFD_NONBLOCK|EFD_SEMAPHORE);
    loop->msg_fifo = async_fifo_new();
    loop->fd_list = NULL;
    loop->timeout_queue = queue_new();
    loop->run_thread = 0;
    loop->quit = 0;
    //eventloop_add_fd(loop,loop->wakeup_fd,EVENTLOOP_IN,on_wakeup,NULL);
    eventloop_add_fd(loop,async_fifo_fd(loop->msg_fifo),EVENTLOOP_IN,(EventloopFdFunc)eventloop_new_cb,loop);
    return loop;
}


void eventloop_free(Eventloop* loop){
    //close(loop->wakeup_fd);
    /* FIXME!! free all data structures */
    async_fifo_free(loop->msg_fifo);
    queue_free(loop->timeout_queue);
    u_free(loop);
}

#endif /* EVENTLOOP_USE_GLIB */


#ifdef EVENTLOOP_USE_GLIB
static gboolean eventloop_add_fd_cb(GSource* src,GIOCondition cond,void* userdata){
    FdData* data = (FdData*)userdata;
    int events = (cond&G_IO_IN?EVENTLOOP_IN:0)|(cond&G_IO_OUT?EVENTLOOP_OUT:0);
    if(data->func(data->userdata,data->fd,events))
        return FALSE;
    return TRUE;
}
#endif
void eventloop_add_fd(Eventloop* loop,int fd,int events,EventloopFdFunc func,void* userdata){
    FdData* data = u_new(FdData,1);
    data->func = func;
    data->userdata = userdata;
    data->fd = fd;
#ifdef EVENTLOOP_USE_GLIB
    {
        GMainContext* c = g_main_loop_get_context(loop);
        GIOChannel* io = g_io_channel_unix_new(fd);
        GIOCondition cond = (events&EVENTLOOP_IN?G_IO_IN:0)|(events&EVENTLOOP_OUT?G_IO_OUT:0);
        GSource* src = g_io_create_watch(io,cond);
        g_source_set_callback(src,(GSourceFunc)eventloop_add_fd_cb,data,u_free);
        g_source_attach(src,c);
    }
#else
    data->events = events;
    if(loop->run_thread && loop->run_thread != thread_self()){
        /* the eventloop is running in another thread */
        Msg* msg = u_new(Msg,1);
        msg->type = MSG_ADD_FD;
        msg->fd_data = data;
        async_fifo_push(loop->msg_fifo,msg);
    }else{
        loop->fd_list = list_prepend(loop->fd_list,data);
    }
#endif
}


#ifdef EVENTLOOP_USE_GLIB
static gboolean eventloop_add_timeout_cb(void* userdata){
    TimeoutData* data = (TimeoutData*)userdata;
    if(data->func(data->userdata)){
        return G_SOURCE_CONTINUE;
    }else{
        return G_SOURCE_REMOVE;
    }
}
#endif
void eventloop_add_timeout(Eventloop* loop,int timeout,EventloopTimeoutFunc func,void* userdata){
    TimeoutData* d = u_new(TimeoutData,1);
    d->func = func;
    d->userdata = userdata;
#ifdef EVENTLOOP_USE_GLIB
    {
        GMainContext* c = g_main_loop_get_context(loop);
        GSource* src = g_timeout_source_new(timeout);
        g_source_set_callback(src,eventloop_add_timeout_cb,d,u_free);
        g_source_attach(src,c);
    }
#else
    d->timeout = timeout;
    d->monotime = get_monotime_now()+timeout;
    if(loop->run_thread && loop->run_thread != thread_self()){
        /* the eventloop is running in another thread */
        Msg* msg = u_new(Msg,1);
        msg->type = MSG_ADD_TIMEOUT;
        msg->timeout_data = d;
        async_fifo_push(loop->msg_fifo,msg);
    }else{
        eventloop_insert_timeout_data(loop,d);
    }
#endif
}


int eventloop_run(Eventloop* loop){
#ifdef EVENTLOOP_USE_GLIB
    g_main_loop_run(loop);
#else
    loop->run_thread = thread_self();
    loop->quit = 0;
    while(!loop->quit){
        fd_set fdread;
        fd_set fdwrite;
        List* iter;
        int maxfd;
        long now;
        long timeout_ms;
        struct timeval timeout;
        int res;

        FD_ZERO(&fdread);
        FD_ZERO(&fdwrite);
        maxfd = -1;
        for(iter=loop->fd_list;iter;iter=iter->next){
            FdData* d = (FdData*)iter->data;
            if(d->events&EVENTLOOP_IN)
                FD_SET(d->fd,&fdread);
            if(d->events&EVENTLOOP_OUT)
                FD_SET(d->fd,&fdwrite);
            if(d->fd>maxfd)
                maxfd = d->fd;
        }

        if(queue_length(loop->timeout_queue)>0){
            timeout_ms = ((TimeoutData*)loop->timeout_queue->head->data)->monotime-get_monotime_now();
            if(timeout_ms<0){
                DEBUG_p("timeout has passed");
                timeout_ms = 0;
            }
            DEBUG_p("timeout in %ld ms",timeout_ms);
            timeout.tv_sec = timeout_ms/1000;
            timeout.tv_usec = (timeout_ms%1000)*1000;
        }else{
            timeout_ms = -1;
        }

        res = select(maxfd+1,&fdread,&fdwrite,NULL,timeout_ms>=0?&timeout:NULL);
        /* If select() returns due to a signal res is <= 0 and the fd-sets
           might be invalid. */
        if(res>0)
            for(iter=loop->fd_list;iter;){
                FdData* d = (FdData*)iter->data;
                List* next = iter->next;
                if(FD_ISSET(d->fd,&fdread)){
                    if(!d->func(d->userdata,d->fd,EVENTLOOP_IN)){
                        loop->fd_list = list_remove(loop->fd_list,iter);
                        u_free(d);
                    }
                }else if(FD_ISSET(d->fd,&fdwrite)){
                    if(!d->func(d->userdata,d->fd,EVENTLOOP_OUT)){
                        loop->fd_list = list_remove(loop->fd_list,iter);
                        u_free(d);
                    }
                }
                iter = next;
            }

        now = get_monotime_now();
        while(queue_length(loop->timeout_queue)>0){
            TimeoutData* d = (TimeoutData*)loop->timeout_queue->head->data;
            if(d->monotime>now)
                break;
            queue_pop_head(loop->timeout_queue);
            if(d->func(d->userdata)){
                d->monotime += d->timeout;
                eventloop_insert_timeout_data(loop,d);
            }else{
                u_free(d);
            }
        }
    }
#endif
    return 0;
}


void eventloop_quit(Eventloop* loop){
#ifdef EVENTLOOP_USE_GLIB
    g_main_loop_quit(loop);
#else
    if(loop->run_thread && loop->run_thread != thread_self()){
        /* the eventloop is running in another thread */
        Msg* msg = u_new(Msg,1);
        msg->type = MSG_QUIT;
        async_fifo_push(loop->msg_fifo,msg);
    }else{
        loop->quit = 1;
    }
#endif
}



int eventloop_other_thread(Eventloop* loop){
#ifdef EVENTLOOP_USE_GLIB
    return 0;
#else
    return loop->run_thread && loop->run_thread != thread_self();
#endif
}


void eventloop_dispatch(Eventloop* loop,DispatchFunc func,void* data){
#ifdef EVENTLOOP_USE_GLIB
    /* FIXME!! */
#else
    if(eventloop_other_thread(loop)){
        Msg* msg = u_new(Msg,1);
        DispatchData* d = u_new(DispatchData,1);
        d->func = func;
        d->data = data;
        msg->type = MSG_DISPATCH;
        msg->dispatch_data = d;
        async_fifo_push(loop->msg_fifo,msg);
    }else{
        DEBUG_p("running right now");
        func(data);
    }
#endif
}


typedef struct{
    Worker inh;
    Eventloop* loop;
}WorkerEventloop;


static void worker_eventloop_done(Worker* d){
    /* nothing */
}


static void worker_eventloop_dispatch(Worker* w,DispatchFunc func,void* data){
    WorkerEventloop* we = (WorkerEventloop*)w;
    eventloop_dispatch(we->loop,func,data);
}


static int worker_eventloop_try_dispatch(Worker* w,DispatchFunc func,void* data){
    WorkerEventloop* we = (WorkerEventloop*)w;
    eventloop_dispatch(we->loop,func,data);
    return 1;
}


static void worker_eventloop_wait(Worker* w){
    /* FIXME!! */
}


Worker* eventloop_create_worker(Eventloop* loop){
    static const WorkerInterface worker_eventloop_vt = {
        .done = worker_eventloop_done,
        .dispatch = worker_eventloop_dispatch,
        .try_dispatch = worker_eventloop_try_dispatch,
        .wait = worker_eventloop_wait
    };
    WorkerEventloop* we = u_new(WorkerEventloop,1);
    ((Worker*)we)->vt = &worker_eventloop_vt;
    we->loop = loop;
    return (Worker*)we;
}
