/*
 * Copyright (C) 2018-2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <limits.h>
#include <stdio.h>
#include <sys/inotify.h>
#include <unistd.h>

#include "filemon.h"
#include "mainloop.h"
#include "memutil.h"

#define BUFFER_SIZE 16

#ifdef USE_DEBUG
static void display_event(struct inotify_event* i){
    printf("    wd =%2d; ", i->wd);
    if (i->cookie > 0)
        printf("cookie =%4d; ", i->cookie);

    printf("mask = ");
    if (i->mask & IN_ACCESS)        printf("IN_ACCESS ");
    if (i->mask & IN_ATTRIB)        printf("IN_ATTRIB ");
    if (i->mask & IN_CLOSE_NOWRITE) printf("IN_CLOSE_NOWRITE ");
    if (i->mask & IN_CLOSE_WRITE)   printf("IN_CLOSE_WRITE ");
    if (i->mask & IN_CREATE)        printf("IN_CREATE ");
    if (i->mask & IN_DELETE)        printf("IN_DELETE ");
    if (i->mask & IN_DELETE_SELF)   printf("IN_DELETE_SELF ");
    if (i->mask & IN_IGNORED)       printf("IN_IGNORED ");
    if (i->mask & IN_ISDIR)         printf("IN_ISDIR ");
    if (i->mask & IN_MODIFY)        printf("IN_MODIFY ");
    if (i->mask & IN_MOVE_SELF)     printf("IN_MOVE_SELF ");
    if (i->mask & IN_MOVED_FROM)    printf("IN_MOVED_FROM ");
    if (i->mask & IN_MOVED_TO)      printf("IN_MOVED_TO ");
    if (i->mask & IN_OPEN)          printf("IN_OPEN ");
    if (i->mask & IN_Q_OVERFLOW)    printf("IN_Q_OVERFLOW ");
    if (i->mask & IN_UNMOUNT)       printf("IN_UNMOUNT ");
    printf("\n");

    if (i->len > 0)
        printf("        name = %s\n", i->name);
}
#endif


typedef struct{
    FilemonFunc func;
    void* userdata;
}FilemonData;


static int on_event(void* userdata,int fd,int fdevent){
    FilemonData* d = (FilemonData*)userdata;
    struct inotify_event buf[BUFFER_SIZE];
    ssize_t bytes;
    int i;

    bytes = read(fd,buf,sizeof(struct inotify_event)*BUFFER_SIZE);
    if(bytes<=0)
        goto stop;

    for(i=0;i*sizeof(struct inotify_event)<bytes;++i){
#ifdef USE_DEBUG
        display_event(&buf[i]);
#endif
        if(!d->func(d->userdata))
            goto stop;
    }
    return 1;
stop:
    u_free(d);
    close(fd);
    return 0;
}


int eventloop_add_filemon(Eventloop* loop,const char* filename,FilemonFunc func,void* data){
    FilemonData* d;
    int fd = inotify_init();
    int wd = inotify_add_watch(fd,filename,IN_MODIFY);
    if(wd<0)
        return wd;

    d = u_new(FilemonData,1);
    d->func = func;
    d->userdata = data;

    eventloop_add_fd(loop,fd,EVENTLOOP_IN,on_event,d);
    return 0;
}


int mainloop_add_filemon(const char* filename,FilemonFunc func,void* data){
    return eventloop_add_filemon(mainloop_get_eventloop(),filename,func,data);
}
