/*
 * Copyright (C) 2010-2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <ctype.h>                      /* isdigit() */
#include <stdlib.h>
#include <sys/select.h>
#include <termios.h>
#include <unistd.h>

#include "debug.h"
#include "keyboard.h"
#include "keyboard-term.h"
#include "mainloop.h"

#define ANSI_TIMEOUT 100

static struct termios term_old;


static int poll_timeout(int fd,int timeout_ms){
    fd_set fdread;
    struct timeval timeout;
    FD_ZERO(&fdread);
    FD_SET(fd,&fdread);
    timeout.tv_sec = timeout_ms/1000;
    timeout.tv_usec = (timeout_ms%1000)*1000;
    return select(fd+1,&fdread,NULL,NULL,&timeout) > 0;
}


static void do_string(const char* str,int len){
    while(len-->0)
        keyboard_key(*str++);
}


static int on_input(void* userdata,int fd,int fdevent){
    static char key[8];
    (void)(1+read(0,key,1));
    switch(*key){
    case '\e': /* escape was pressed */
        if(poll_timeout(0,ANSI_TIMEOUT)){
            /* something more was pressed */
            enum {GOT_ESC,GOT_ESCH /* ,GOT_ESCHH,GOT_WAITTILDE */} state = GOT_ESC;;
            int param = 0;
            int len = 1;
            for(;;){
                /* keep the chars in a buffer if the code breaks */
                (void)(1+read(0,&key[len],1));
                DEBUG_p("got '%c' (%d)",key[len],key[len]);
                switch(state){
                case GOT_ESC:
                    switch(key[len]){
                    case '[':
                        ++len;
                        state = GOT_ESCH;
                        continue;
                    default:
                        /* something other than '[' after escape: code broken */
                        DEBUG_p("code broken after escape");
                        do_string(key,len+1);
                        return 1;
                    }
                    break;
                case GOT_ESCH:
                    switch(key[len]){
#if 0
                    case '[':
                        state = GOT_ESCHH;
                        continue;
#endif
                    case '~':
                        switch(param){
                        case 5:
                            keyboard_key(KEYBOARD_PAGEUP);
                            return 1;
                        case 6:
                            keyboard_key(KEYBOARD_PAGEDOWN);
                            return 1;
                        default:
                            DEBUG_p("unknown param before ~: %d",param);
                            //do_string(key,len+1);
                            return 1;
                        }
                        break;
                    case 'A': /* cursor up */
                        keyboard_key(KEYBOARD_UP);
                        return 1;
                    case 'B': /* cursor down */
                        keyboard_key(KEYBOARD_DOWN);
                        return 1;
                    case 'C': /* cursor right */
                        keyboard_key(KEYBOARD_RIGHT);
                        return 1;
                    case 'D': /* cursor left */
                        keyboard_key(KEYBOARD_LEFT);
                        return 1;
                    case 'H':
                        keyboard_key(KEYBOARD_HOME);
                        return 1;
                    case 'F':
                        keyboard_key(KEYBOARD_END);
                        return 1;
                    default:
                        if(isdigit(key[len])){
                            param *= 10;
                            param += key[len]-'0';
                            ++len;
                            continue;
                        }
                        /* code broken */
                        DEBUG_p("code broken after '[' (%c)",key[len]);
                        do_string(key,len+1);
                        return 1;
                    }
                    break;
                }
            }
        }
        break;
    default:
        break;
    }
    keyboard_key(*key);
    return 1;
}


static void restore_terminal(){
    tcsetattr(0,TCSANOW,&term_old);
}


int keyboard_term_init(){
    struct termios term_new;
    if(tcgetattr(0,&term_old))
        return -1;
    atexit(restore_terminal);
    term_new = term_old;
    term_new.c_lflag &= ~ECHO;          /* disable echo */
    term_new.c_lflag &= ~ICANON;        /* uncanonize */
    if(tcsetattr(0,TCSANOW,&term_new))
        return -1;
    mainloop_add_fd(0,EVENTLOOP_IN,on_input,NULL);
    return 0;
}


void keyboard_term_done(){
    restore_terminal();
}
