/*
 * Copyright (C) 2018-2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "config.h"

#include <abydos.h>
#include <cairo.h>
#include <cairo-xlib.h>
#include <extractor.h>
#include <getopt.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>                  /* XParseGeometry() */

#include "debug.h"
#include "filemon.h"
#include "keyboard.h"
#include "keyboard-term.h"
#include "keyboard-xlib.h"
#include "mainloop.h"
#include "memutil.h"

#ifndef ABS
# define ABS(a) ((a)<0?-(a):(a))
#endif
#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#define _(a) (a)

#define MOUSE_SCALE 1.25
#define KEYBOARD_SCALE 1.25
#define KEYBOARD_PANE 16
#define CHECKER_SIZE 8

typedef enum{
    BG_BLACK,
    BG_WHITE,
    BG_MAGENTA,
    BG_PATTERN
}background_t;

typedef enum{
    Z_SET,
    Z_FIT
}zoom_t;

static struct EXTRACTOR_PluginList* plugins;

const char* progname;
static const char* opt_mime_type;       /* Forced MIME type. */
static int opt_version;
static const char* filename;
static char* mime_type;                 /* Detected MIME type. */

static int gravity_right;
static int gravity_down;
static int winx = -1;
static int winy = -1;
static int winw = -1;
static int winh = -1;
static cairo_surface_t* surface;
static abydos_t* ar;
static int page;
static abydos_info_t info;
static double animtime;

static Display* disp;
static int scrn;
#ifdef X_HAVE_UTF8_STRING
static XIM xim;
static XIC xic;
#endif

static cairo_pattern_t* pattern;

static background_t background;
static zoom_t zoom = Z_FIT;
static double scale = 1;
static int panex;
static int paney;
static int rot;
static int ancx;
static int ancy;

static int dirty;


static void draw(){
    cairo_t* cr;
    int imgw;
    int imgh;

    if(!disp)
        return;
    if(!surface)
        return;

    DEBUG_p("dirty = %d",dirty);

    cr = cairo_create(surface);
    imgw = info.width;
    imgh = info.height;

    if(rot&1){
        int tmp = imgw;
        imgw = imgh;
        imgh = tmp;
    }

    switch(zoom){
    case Z_SET:
        break;
    case Z_FIT:
        scale = MIN((double)winw/imgw,(double)winh/imgh);
        break;
    }

    cairo_push_group(cr);
    switch(background){
    case BG_BLACK:
        cairo_set_source_rgb(cr,0,0,0);
        break;
    case BG_WHITE:
        cairo_set_source_rgb(cr,1,1,1);
        break;
    case BG_MAGENTA:
        cairo_set_source_rgb(cr,1,0,1);
        break;
    case BG_PATTERN:
        if(!pattern){
            cairo_surface_t* surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,CHECKER_SIZE*2,CHECKER_SIZE*2);
            cairo_t* cr2 = cairo_create(surface);
            cairo_set_source_rgb(cr2,2.0/3,2.0/3,2.0/3);
            cairo_rectangle(cr2,0,0,CHECKER_SIZE,CHECKER_SIZE);
            cairo_rectangle(cr2,CHECKER_SIZE,CHECKER_SIZE,CHECKER_SIZE,CHECKER_SIZE);
            cairo_fill(cr2);
            cairo_set_source_rgb(cr2,1.0/3,1.0/3,1.0/3);
            cairo_rectangle(cr2,CHECKER_SIZE,0,CHECKER_SIZE,CHECKER_SIZE);
            cairo_rectangle(cr2,0,CHECKER_SIZE,CHECKER_SIZE,CHECKER_SIZE);
            cairo_fill(cr2);
            cairo_destroy(cr2);
            pattern = cairo_pattern_create_for_surface(surface);
            cairo_surface_destroy(surface);
            cairo_pattern_set_extend(pattern,CAIRO_EXTEND_REPEAT);
        }
        cairo_set_source(cr,pattern);
        break;
    }
    cairo_paint(cr);

    if(winw>=imgw*scale){
        panex = (winw-imgw*scale)/2;
    }else if(panex>0){
        panex = 0;
    }else if(panex<winw-imgw*scale){
        panex = winw-imgw*scale;
    }
    if(winh>=imgh*scale){
        paney = (winh-imgh*scale)/2;
    }else if(paney>0){
        paney = 0;
    }else if(paney<winh-imgh*scale){
        paney = winh-imgh*scale;
    }

    cairo_translate(cr,panex,paney);
    cairo_scale(cr,scale,scale);
    cairo_translate(cr,imgw/2,imgh/2);
    cairo_rotate(cr,rot*M_PI/2);
    cairo_translate(cr,-info.width/2,-info.height/2);

    abydos_render(ar,cr,0);
    cairo_pop_group_to_source(cr);
    cairo_paint(cr);
    cairo_destroy(cr);
    cairo_surface_flush(surface);
    XFlush(disp);
    dirty = 0;
}


static void set_page(int nr){
    if(ar){
        abydos_set_page(ar,nr);
        page = abydos_get_page(ar);
        ++dirty;
    }else{
        page = nr;
    }
}


static void set_scale(double s,int x,int y){
    double offs;
    if(x<0)
        x = winw/2;
    if(y<0)
        y = winh/2;
    offs = x-panex;
    offs *= s/scale;
    panex = x-offs;
    offs = y-paney;
    offs *= s/scale;
    paney = y-offs;
    zoom = Z_SET;
    scale = s;
    ++dirty;
}


static void set_background(background_t bg){
    background = bg;
    ++dirty;
}


static double get_monotime(){
    struct timespec monotime;
    clock_gettime(CLOCK_MONOTONIC,&monotime);
    return monotime.tv_sec+(double)monotime.tv_nsec/1000000000;
}


static int on_animate(void* userdata){
    double now = get_monotime();
    abydos_set_time(ar,abydos_get_time(ar) + now - animtime);
    animtime = now;
    mainloop_add_timeout(ceil(abydos_get_delay(ar)*1000),on_animate,NULL);
    draw();
    return 0;
}


static int on_extract(
    void* userdata,
    const char* plugin_name,
    enum EXTRACTOR_MetaType type,
    enum EXTRACTOR_MetaFormat format,
    const char *data_mime_type,
    const char *data,
    size_t data_len
){
    if(EXTRACTOR_METATYPE_MIMETYPE==type){
        mime_type = u_strndup(data,data_len);
        return 1;
    }
    return 0;
}


static int load(void){
    if(opt_mime_type){
        mime_type = u_strdup(opt_mime_type);
    }else{
        EXTRACTOR_extract(plugins,filename,NULL,0,on_extract,&mime_type);
        if(!mime_type){
            fprintf(stderr,"Could not determine MIME type of %s\n",filename);
            return 0;
        }
        DEBUG_p("mime type: %s",mime_type);
    }

    ar = abydos_create_from_file(mime_type,filename);
    if(!ar){
        fprintf(stderr,"Failed to open %s (with MIME type %s)\n",filename,mime_type);
        return 0;
    }
    abydos_get_info(ar,&info);

    set_page(page);
    if(abydos_get_frame_count(ar)>1){
        animtime = get_monotime();
        mainloop_add_timeout(ceil(abydos_get_delay(ar)*1000),on_animate,NULL);
    }

    return 1;
}


static int on_reload(void* userdata){
    if(mime_type){
        u_free(mime_type);
        mime_type = NULL;
    }
    if(ar){
        abydos_destroy(ar);
        ar = NULL;
    }
    load();
    return 1;
}


static int operation(int op){
    switch(op){
    case 'b':
        set_background(BG_BLACK);
        return 1;
    case 'w':
        set_background(BG_WHITE);
        return 1;
    case 'm':
        set_background(BG_MAGENTA);
        return 1;
    case 'p':
        set_background(BG_PATTERN);
        return 1;
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
        set_scale(op-'0',-1,-1);
        return 1;
    case 'x':
        zoom = Z_FIT;
        ++dirty;
        return 1;
    case 'r':
        if(++rot>=4)
            rot = 0;
        ++dirty;
        return 1;
    }
    return 0;
}


static int on_key(void* userdata,int key){
    if(operation(key))
        return 1;
    switch(key){
    case '\e':
    case 'q':
        mainloop_quit();
        return 1;
    case KEYBOARD_HOME:
        set_page(0);
        return 1;
    case KEYBOARD_END:
        set_page(abydos_get_page_count(ar)-1);
        return 1;
    case ' ':
    case KEYBOARD_PAGEDOWN:
        set_page(page+1);
        return 1;
    case '\b':
    case KEYBOARD_PAGEUP:
        set_page(page-1);
        return 1;
    case '+':
        set_scale(scale * KEYBOARD_SCALE,-1,-1);
        break;
    case '-':
        set_scale(scale / KEYBOARD_SCALE,-1,-1);
        break;
    case KEYBOARD_UP:
        paney += KEYBOARD_PANE;
        ++dirty;
        return 1;
    case KEYBOARD_DOWN:
        paney -= KEYBOARD_PANE;
        ++dirty;
        return 1;
    case KEYBOARD_LEFT:
        panex += KEYBOARD_PANE;
        ++dirty;
        return 1;
    case KEYBOARD_RIGHT:
        panex -= KEYBOARD_PANE;
        ++dirty;
        return 1;
    }
    return 0;
}


static int on_xevent(void* userdata,int fd,int fdevent){
    DEBUG_p("begin");
    while(XPending(disp)){
        XEvent ev;
        XNextEvent(disp,&ev);
        switch(ev.type){
        case ClientMessage:
            mainloop_quit();
            break;
        case Expose:
            ++dirty;
            break;
        case ConfigureNotify:
            winw = ev.xconfigure.width;
            winh = ev.xconfigure.height;
            DEBUG_p("resize to %dx%d",winw,winh);
            cairo_xlib_surface_set_size(surface,winw,winh);
            break;
        case KeyPress:
            keyboard_xlib_key(&ev.xkey,xic);
            break;
        case ButtonPress:
            DEBUG_p("button press: %d",ev.xbutton.button);
            switch(ev.xbutton.button){
            case 1:
            case 2:
            case 3:
                /* set pane ancor */
                ancx = ev.xbutton.x;
                ancy = ev.xbutton.y;
                break;
            case 4:
                /* zoom in */
                set_scale(scale * MOUSE_SCALE,ev.xbutton.x,ev.xbutton.y);
                break;
            case 5:
                /* zoom out */
                set_scale(scale / MOUSE_SCALE,ev.xbutton.x,ev.xbutton.y);
                break;
            }
            break;
        case MotionNotify:
            panex += ev.xmotion.x - ancx;
            paney += ev.xmotion.y - ancy;
            ancx = ev.xmotion.x;
            ancy = ev.xmotion.y;
            ++dirty;
            break;
        }
    }
    if(dirty)
        draw();
    return 1;
}


static void usage(int on_stderr){
    FILE* f = on_stderr?stderr:stdout;
    fprintf(f,_("Usage: %s [options] <image>\n"
                "\n"
                "Options:\n"),progname);
    fprintf(f,_("  -h --help             Show this help message and exit\n"));
    fprintf(f,_("  -P --page <PAGE>      Start at page <PAGE>\n"));
    fprintf(f,_("  -b --black            Black background (the default)\n"));
    fprintf(f,_("  -w --white            White background\n"));
    fprintf(f,_("  -m --magenta          Magenta background\n"));
    fprintf(f,_("  -p --pattern          Checkered pattern packground\n"));
    fprintf(f,_("  -1 ... -8             Set zoom\n"));
    fprintf(f,_("  -x --fit              Scale the image to fit the window\n"));
    fprintf(f,_("  -r --rotate           Rotate the image clockwise (can be given multiple times)\n"));
    fprintf(f,_("  -M --mime-type <TYPE> Force the image to be interpreted as <TYPE>\n"));
    fprintf(f,_("  -G --geometry <GEOM>  Set window size and/or location to <GEOM>\n"));
    fprintf(f,_("  -V --version          Display version (" VERSION ") and exit\n"));
}


int main(int argc,char** argv) {
    Window w;
    int x11_fd;

    for(;;){
        static const struct option entries[] = {
            {"help",no_argument,NULL,'h'},
            {"page",required_argument,NULL,'P'},
            {"black",no_argument,NULL,'b'},
            {"white",no_argument,NULL,'w'},
            {"pattern",no_argument,NULL,'p'},
            {"fit",no_argument,NULL,'x'},
            {"rotate",no_argument,NULL,'r'},
            {"mime-type",required_argument,NULL,'M'},
            {"geometry",required_argument,NULL,'G'},
            {"version",no_argument,NULL,'V'},
            {NULL}
        };
        int index;
        int c = getopt_long(argc,argv,"hP:bwp12345678xrM:G:V",entries,&index);
        if(c<0)
            break;
        if(operation(c))
            continue;
        switch(c){
        case 'h':
            usage(0);
            return 0;
        case 'P':
            set_page(atoi(optarg));
            break;
        case 'M':
            opt_mime_type = optarg;
            break;
        case 'G':
            {
                int x,y;
                unsigned int w,h;
                int flags = XParseGeometry(optarg,&x,&y,&w,&h);
                gravity_down = flags&YNegative;
                gravity_right = flags&XNegative;
                if(flags&XValue)
                    winx = ABS(x);
                if(flags&YValue)
                    winy = ABS(y);
                if(flags&WidthValue)
                    winw = w;
                if(flags&HeightValue)
                    winh = h;
            }
            break;
        case 'V':
            opt_version = 1;
            break;
        default:
            usage(1);
            return 1;
        }
    }
    if(opt_version){
        printf("%s " VERSION "\n",progname);
        return 0;
    }
    if(optind >= argc){
        usage(1);
        return 2;
    }
    filename = argv[optind];

    /* open connection with the server */
    disp = XOpenDisplay(NULL);
    if(!disp){
        fprintf(stderr,"Cannot open display\n");
        return 1;
    }

    plugins = EXTRACTOR_plugin_add_defaults(EXTRACTOR_OPTION_DEFAULT_POLICY);

    mainloop_init();

    if(!load())
        return 2;
    mainloop_add_filemon(filename,on_reload,NULL);

    keyboard_subscribe(on_key,NULL);
    keyboard_term_init();

    scrn = DefaultScreen(disp);
    {
        XWindowAttributes ra;
        XGetWindowAttributes(disp,DefaultRootWindow(disp),&ra);
        if(winw<=0)
            winw = MIN(info.width,ra.width);
        if(winh<=0)
            winh = MIN(info.height,ra.height);
        if(winx<0){
            winx = (ra.width-winw)/2;
        }else if(gravity_right){
            winx = ra.width-winw-winx;
        }
        if(winy<0){
            winy = (ra.height-winh)/2;
        }else if(gravity_down){
            winy = ra.height-winh-winy;
        }
    }
    DEBUG_p("geometry: %dx%d @ %dx%d",winw,winh,winx,winy);
    w = XCreateSimpleWindow(disp,RootWindow(disp,scrn),winx,winy,winw,winh,1,BlackPixel(disp,scrn),WhitePixel(disp,scrn));

    /* Process Window Close Event through event handler so XNextEvent does Not fail */
    Atom delWindow = XInternAtom(disp,"WM_DELETE_WINDOW",0);
    XSetWMProtocols(disp,w,&delWindow,1);
    XSelectInput(disp,w,ExposureMask|StructureNotifyMask|KeyPressMask|ButtonPressMask|ButtonReleaseMask|ButtonMotionMask);
    XStoreName(disp,w,filename);
    XMapWindow(disp,w);

#ifdef X_HAVE_UTF8_STRING
    XSetLocaleModifiers("");
    xim = XOpenIM(disp,0,0,0);
    if(!xim){
        XSetLocaleModifiers("@im=none");
        xim = XOpenIM(disp,0,0,0);
    }
    xic = XCreateIC(xim,XNInputStyle,XIMPreeditNothing|XIMStatusNothing,NULL);
#endif

    surface = cairo_xlib_surface_create(disp,w,DefaultVisual(disp,scrn),100,100);

    x11_fd = ConnectionNumber(disp);
    mainloop_add_fd(x11_fd,EVENTLOOP_IN,on_xevent,disp);
    /* Call the event handler once to get things going. */
    on_xevent(disp,0,0);

    mainloop_run();

#ifdef X_HAVE_UTF8_STRING
    XDestroyIC(xic);
    XCloseIM(xim);
#endif

    XDestroyWindow(disp,w);
    keyboard_term_done();
    mainloop_done();
    EXTRACTOR_plugin_remove_all(plugins);
    XCloseDisplay(disp);

    return 0;
}
