/*
 * Copyright (C) 2015-2018 Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "mainloop.h"

#ifdef MAINLOOP_USE_GLIB
# include <glib.h>
# include <glib-unix.h>
#endif
#include <signal.h>
#include <stdint.h>
#include <unistd.h>

#include "dispatch.h"
#include "memutil.h"


#ifndef POINTER2INT
# ifdef MAINLOOP_USE_GLIB
#  define POINTER2INT(a) GPOINTER_TO_INT(a)
#  define INT2POINTER(a) GINT_TO_POINTER(a)
# else
#  define POINTER2INT(a) ((int)(intptr_t)(a))
#  define INT2POINTER(a) ((void*)(intptr_t)(a))
# endif
#endif

#define SIGMAX 63

typedef void(*sighandler_t)(int);

typedef struct{
    sighandler_t old_handler;
    Signal signal_signal;
}SignalInfo;


static Eventloop* mainloop;
static SignalInfo signal_info[SIGMAX];
static int signal_pipe[2];


static int signal_marshal(SignalHandler handler,void* userdata,void* signaldata){
    int signum = POINTER2INT(signaldata);
    MainloopSignalFunc cb = (MainloopSignalFunc)handler;
    return cb(userdata,signum);
}


/*
static void on_signal_cb(void* data){
    int index = POINTER2INT(data)-1;
    signal_invoke_marshal(&signal_info[index].signal_signal,signal_marshal,data);
}
*/
#ifdef _MAINLOOP_USE_GLIB
static int on_signal(void* userdata){
    int signum = POINTER2INT(userdata);
    signal_invoke_marshal(&signal_info[signum-1].signal_signal,signal_marshal,userdata);
    return G_SOURCE_CONTINUE;
}
#else
static void on_signal(int signum){
    //mainloop_dispatch(on_signal_cb,INT2POINTER(signum));
    write(signal_pipe[1],&signum,sizeof(int));
}
int on_signal_pipe(void* userdata,int fd,int events){
    int signum;
    read(fd,&signum,sizeof(int));
    signal_invoke_marshal(&signal_info[signum-1].signal_signal,signal_marshal,INT2POINTER(signum));
    return 0;
}
#endif



int mainloop_init(){
    int i;
#ifdef MAINLOOP_USE_GLIB
    mainloop = g_main_loop_new(NULL,FALSE);
#else
    mainloop = eventloop_new();
#endif
    for(i=0;i<SIGMAX;++i)
        signal_init(&signal_info[i].signal_signal);
#ifndef _MAINLOOP_USE_GLIB
    pipe(signal_pipe);
    mainloop_add_fd(signal_pipe[0],EVENTLOOP_IN,on_signal_pipe,NULL);
#endif
    return 1;
}


void mainloop_done(){
    /* FIXME!! clean up signal handlers */
    eventloop_free(mainloop);
}



Eventloop* mainloop_get_eventloop(){
    return mainloop;
}



int mainloop_run(){
    return eventloop_run(mainloop);
}


void mainloop_quit(){
    eventloop_quit(mainloop);
}



#ifdef MAINLOOP_USE_GLIB
typedef struct{
    EventloopFdFunc func;
    void* userdata;
}MainloopAddFdData;
static gboolean mainloop_add_fd_cb(int fd,GIOCondition cond,void* userdata){
    MainloopAddFdData* data = (MainloopAddFdData*)userdata;
    int events = (cond&G_IO_IN?EVENTLOOP_IN:0)|(cond&G_IO_OUT?EVENTLOOP_OUT:0);
    if(data->func(data->userdata,fd,events)){
        u_free(data);
        return FALSE;
    }else{
        return TRUE;
    }
}
#endif
void mainloop_add_fd(int fd,int events,EventloopFdFunc func,void* userdata){
#ifdef MAINLOOP_USE_GLIB
    MainloopAddFdData* data = u_new(MainloopAddFdData,1);
    GIOCondition cond = (events&EVENTLOOP_IN?G_IO_IN:0)|(events&EVENTLOOP_OUT?G_IO_OUT:0);
    data->func = func;
    data->userdata = userdata;
    g_unix_fd_add(fd,cond,mainloop_add_fd_cb,data);
#else
    eventloop_add_fd(mainloop,fd,events,func,userdata);
#endif
}


#ifdef MAINLOOP_USE_GLIB
typedef struct{
    EventloopTimeoutFunc func;
    void* userdata;
}AddTimeoutData;
static gboolean mainloop_add_timeout_cb(void* userdata){
    AddTimeoutData* d = (AddTimeoutData*)userdata;
    if(d->func(d->userdata)){
        u_free(d);
        return FALSE;
    }else{
        return TRUE;
    }
}
#endif
void mainloop_add_timeout(int timeout,EventloopTimeoutFunc func,void* userdata){
#ifdef MAINLOOP_USE_GLIB
    AddTimeoutData* d = u_new(AddTimeoutData,1);
    d->func = func;
    d->userdata = userdata;
    g_timeout_add(timeout,mainloop_add_timeout_cb,d);
#else
    eventloop_add_timeout(mainloop,timeout,func,userdata);
#endif
}


SignalSubscription* mainloop_subscribe_signal(int signum,MainloopSignalFunc func,void* userdata){
    int index;
    if((signum<1) || (signum>SIGMAX))
        return NULL;
    index = signum-1;
    if(!signal_info[index].signal_signal.subscription_list){
#ifdef _MAINLOOP_USE_GLIB
        signal_info[index].tag = g_unix_signal_add(signum,on_signal,INT2POINTER(signum));
#else
        signal_info[index].old_handler = signal(signum,on_signal);
#endif
    }
    return signal_subscribe(&signal_info[index].signal_signal,(SignalHandler)func,userdata);
}


void mainloop_unsubscribe_signal(int signum,SignalSubscription* ss){
    int index;
    if(signum<1 || signum>SIGMAX)
        return;
    index = signum-1;
    signal_unsubscribe(&signal_info[index].signal_signal,ss);
    if(!signal_info[signum].signal_signal.subscription_list){
#ifdef _MAINLOOP_USE_GLIB
        g_source_remove(signal_info[index].tag);
#else
        signal(signum+1,signal_info[index].old_handler);
#endif
    }
}



#ifdef MAINLOOP_USE_GLIB
typedef struct{
    DispatchFunc func;
    void* userdata;
}MainloopDispatchData;
static gboolean mainloop_dispatch_cb(void* userdata){
    MainloopDispatchData* data = (MainloopDispatchData*)userdata;
    data->func(data->userdata);
    u_free(data);
    return G_SOURCE_REMOVE;
}
#endif
void mainloop_dispatch(DispatchFunc func,void* userdata){
#ifdef MAINLOOP_USE_GLIB
    MainloopDispatchData* data = u_new(MainloopDispatchData,1);
    data->func = func;
    data->userdata = userdata;
    g_idle_add(mainloop_dispatch_cb,data);
#else
    eventloop_dispatch(mainloop,func,userdata);
#endif
}


static void worker_mainloop_done(Worker* w){
    /* nothing */
}


static void worker_mainloop_dispatch(Worker* w,DispatchFunc func,void* data){
    mainloop_dispatch(func,data);
}


static int worker_mainloop_try_dispatch(Worker* w,DispatchFunc func,void* data){
    mainloop_dispatch(func,data);
    return 1;
}


static void worker_mainloop_wait(Worker* w){
    /* FIXME!! how to handle this? */
}


Worker* mainloop_get_worker(){
    static const WorkerInterface worker_mainloop_vt = {
        .done = worker_mainloop_done,
        .dispatch = worker_mainloop_dispatch,
        .try_dispatch = worker_mainloop_try_dispatch,
        .wait = worker_mainloop_wait
    };
    static const Worker wm = {
        .vt = &worker_mainloop_vt
    };
    return (Worker*)&wm;
}
