/*
 * Copyright (C) 2015 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "queue.h"

#ifndef QUEUE_USE_GLIB

#include "memutil.h"

void queue_free(Queue* queue){
    list_free(queue->head);
    u_free(queue);
}


void queue_free_full(Queue* q,ListNotifyFunc func){
    list_free_full(q->head,func);
    u_free(q);
}


void queue_clear(Queue* q){
    list_free(q->head);
    q->head = q->tail = NULL;
    q->length = 0;
}


void queue_clear_full(Queue* q,ListNotifyFunc func){
    List* iter;
    for(iter=q->head;iter;iter=iter->next)
        func(iter->data);
    queue_clear(q);
}


List* queue_push_head(Queue* q,void* data){
    List* node = u_new(List,1);
    node->prev = NULL;
    node->data = data;
    if(q->length++<=0){
        node->next = NULL;
        q->head = q->tail = node;
    }else{
        node->next = q->head;
        q->head->prev = node;
        q->head = node;
    }
    return node;
}


List* queue_push_tail(Queue* q,void* data){
    List* node = u_new(List,1);
    node->next = NULL;
    node->data = data;
    if(q->length++<=0){
        node->prev = NULL;
        q->head = q->tail = node;
    }else{
        node->prev = q->tail;
        q->tail->next = node;
        q->tail = node;
    }
    return node;
}


void* queue_pop_head(Queue* q){
    List* node;
    void* data;
    if(q->length<=0)
        return NULL;
    node = q->head;
    if(--q->length<=0){
        q->head = q->tail = NULL;
    }else{
        q->head = node->next;
        q->head->prev = NULL;
    }
    data = node->data;
    u_free(node);
    return data;
}


void* queue_pop_tail(Queue* q){
    List* node;
    void* data;
    if(q->length<=0)
        return NULL;
    node = q->tail;
    if(--q->length<=0){
        q->head = q->tail = NULL;
    }else{
        q->tail = node->prev;
        q->tail->next = NULL;
    }
    data = node->data;
    u_free(node);
    return data;
}


List* queue_insert_before(Queue* q,List* iter,void* data){
    if(iter->prev){
        List* node = u_new(List,1);
        node->prev = iter->prev;
        node->next = iter;
        node->data = data;
        node->prev->next = node;
        iter->prev = node;
        return node;
    }else{
        return queue_push_head(q,data);
    }
}


List* queue_insert_after(Queue* q,List* iter,void* data){
    if(iter->next){
        List* node = u_new(List,1);
        node->prev = iter;
        node->next = iter->next;
        node->data = data;
        node->next->prev = node;
        iter->next = node;
        return node;
    }else{
        return queue_push_tail(q,data);
    }
}


void* queue_remove(Queue* q,List* iter){
    if(!iter->prev){
        return queue_pop_head(q);
    }else if(!iter->next){
        return queue_pop_tail(q);
    }else{
        void* data = iter->data;
        iter->prev->next = iter->next;
        iter->next->prev = iter->prev;
        u_free(iter);
        return data;
    }
}


List* queue_find_data(Queue* q,void* data){
    List* iter;
    for(iter=q->tail;iter;iter=iter->prev)
        if(iter->data==data)
            return iter;
    return NULL;
}


List* queue_find_custom(Queue* q,void* data,QueueCompFunc cmp){
#if 1
    /* this returns the iter for the first matching member */
    List* iter;
    for(iter=q->tail;iter;iter=iter->prev)
        if(!cmp(iter->data,data))
            return iter;
    return NULL;
#else
    /* this returns a list of all matching members */
    List* list = NULL;
    List* iter;
    for(iter=q->tail;iter;iter=iter->prev)
        if(!cmp(iter->data,data))
            list = list_prepend(list,iter->data);
    return list;
#endif
}

#endif /* !USE_GLIB */
