/*
 * Copyright (C) 2015 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "debug.h"
#include "list.h"
#include "memutil.h"
#include "signals.h"


#ifdef USE_REFCOUNT
static SignalSubscription* signal_subscription_new(SignalHandler handler,void* userdata,SignalUserdataDestroy destroy){
#else
static SignalSubscription* signal_subscription_new(SignalHandler handler,void* userdata){
#endif
    SignalSubscription* subscription;
    subscription = u_new(SignalSubscription,1);
    subscription->handler = handler;
    subscription->userdata = userdata;
#ifdef USE_REFCOUNT
    subscription->destroy = destroy;
#endif
    return subscription;
}


#ifdef USE_REFCOUNT
static void signal_subscription_free(SignalSubscription* subscription){
    if(subscription->destroy)
        subscription->destroy(subscription->userdata);
    u_free(subscription);
}
#else
#define signal_subscription_free u_free
#endif



Signal* signal_init(Signal* signal){
    signal->subscription_list = NULL;
    return signal;
}


void signal_done(Signal* signal){
    List* iter;
    /* Free all subscribers. */
    for(iter=signal->subscription_list;iter;iter=iter->next){
        SignalSubscription* subscription;
        subscription = (SignalSubscription*)iter->data;
        signal_subscription_free(subscription);
    }
    list_free(signal->subscription_list);
}


Signal* signal_new(){
    return signal_init(u_new(Signal,1));
}


void signal_free(Signal* signal){
    signal_done(signal);
    u_free(signal);
}



#ifdef USE_REFCOUNT
SignalSubscription* signal_subscribe_full(Signal* signal,SignalHandler handler,void* userdata,SignalUserdataDestroy destroy){
#else
SignalSubscription* signal_subscribe(Signal* signal,SignalHandler handler,void* userdata){
#endif
    SignalSubscription* subscription;
#ifdef USE_REFCOUNT
    subscription = signal_subscription_new(handler,userdata,destroy);
#else
    subscription = signal_subscription_new(handler,userdata);
#endif
    signal->subscription_list = list_prepend(signal->subscription_list,subscription);
    return subscription;
}


#ifdef USE_REFCOUNT
SignalSubscription* signal_subscribe_last_full(Signal* signal,SignalHandler handler,void* userdata,SignalUserdataDestroy destroy){
#else
SignalSubscription* signal_subscribe_last(Signal* signal,SignalHandler handler,void* userdata){
#endif
    SignalSubscription* subscription;
#ifdef USE_REFCOUNT
    subscription = signal_subscription_new(handler,userdata,destroy);
#else
    subscription = signal_subscription_new(handler,userdata);
#endif
    signal->subscription_list = list_append(signal->subscription_list,subscription);
    return subscription;
}


#ifdef USE_REFCOUNT
void signal_unsubscribe(Signal* signal,SignalSubscription* subscription){
#else
void* signal_unsubscribe(Signal* signal,SignalSubscription* subscription){
    void* userdata;
    userdata = signal_subscription_get_userdata(subscription);
#endif
    signal->subscription_list = list_remove_data(signal->subscription_list,subscription);
    signal_subscription_free(subscription);
#ifndef USE_REFCOUNT
    return userdata;
#endif
}



int signal_invoke(Signal* signal,void* signaldata){
    List* iter;
    for(iter=signal->subscription_list;iter;){
        SignalSubscription* subscription;
        int result;
        subscription = (SignalSubscription*)iter->data;
        /* advance the iter early if the signal handler unsubscribes the signal */
        iter=iter->next;
        if((result = subscription->handler(subscription->userdata,signaldata)))
            return result;
    }
    return 0;
}


int signal_invoke_marshal(Signal* signal,SignalMarshall marshall,void* signaldata){
    List* iter;
    for(iter=signal->subscription_list;iter;){
        SignalSubscription* subscription;
        int result;
        subscription = (SignalSubscription*)iter->data;
        /* advance the iter early if the signal handler unsubscribes the signal */
        iter=iter->next;
        //DEBUG_p("handler: %p",subscription->handler);
        if((result = marshall(subscription->handler,subscription->userdata,signaldata)))
            return result;
    }
    return 0;
}
